@@ -3,7 +3,7 @@
 '''
 A library to access Tryton's models like a client.
 '''
-__version__ = "3.0.2"
+__version__ = "3.2.0"
 __all__ = ['Model', 'Wizard']
 import sys
 try:
@@ -31,7 +31,7 @@ class _EvalEnvironment(dict):
         self.parent = parent
 
     def __getitem__(self, item):
-        if item == '_parent_' + self.parent._parent_name \
+        if item == '_parent_' + self.parent._parent_field_name \
                 and self.parent.parent:
             return _EvalEnvironment(self.parent.parent)
         return self.parent._get_eval()[item]
@@ -143,6 +143,8 @@ class NumericDescriptor(FieldDescriptor):
 class ReferenceDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
         value = super(ReferenceDescriptor, self).__get__(instance, owner)
+        if instance._parent_name == self.name:
+            value = instance._parent
         if isinstance(value, basestring):
             model_name, id = value.split(',', 1)
             if model_name:
@@ -196,6 +198,8 @@ class Many2OneDescriptor(FieldDescriptor):
     def __get__(self, instance, owner):
         relation = Model.get(self.definition['relation'], instance._config)
         value = super(Many2OneDescriptor, self).__get__(instance, owner)
+        if instance._parent_name == self.name:
+            value = instance._parent
         if isinstance(value, (int, long)) and value is not False:
             value = relation(value)
         elif not value:
@@ -272,25 +276,27 @@ class One2OneValueDescriptor(Many2OneValueDescriptor):
 
 class One2ManyValueDescriptor(ValueDescriptor):
     def __get__(self, instance, owner):
-        value = [('add', [])]
+        value = []
         value_list = getattr(instance, self.name)
-        parent_name = self.definition.get('relation_field', '')
+        to_add = []
         to_create = []
+        to_write = []
         for record in value_list:
             if record.id >= 0:
-                values = record._get_values(fields=record._changed)
-                values.pop(parent_name, None)
-                if record._changed and values:
-                    value.append(('write', [record.id], values))
-                value[0][1].append(record.id)
+                if record._changed:
+                    to_write.extend(([record.id],
+                            record._get_values(fields=record._changed)))
+                to_add.append(record.id)
             else:
-                values = record._get_values()
-                values.pop(parent_name, None)
-                to_create.append(values)
+                to_create.append(record._get_values())
+        if to_add:
+            value.append(('add', to_add))
         if to_create:
             value.append(('create', to_create))
+        if to_write:
+            value.append(('write',) + tuple(to_write))
         if value_list.record_removed:
-            value.append(('unlink', [x.id for x in value_list.record_removed]))
+            value.append(('remove', [x.id for x in value_list.record_removed]))
         if value_list.record_deleted:
             value.append(('delete', [x.id for x in value_list.record_deleted]))
         return value
@@ -344,7 +350,6 @@ class MetaModelFactory(object):
         'boolean': BooleanDescriptor,
         'char': CharDescriptor,
         'text': CharDescriptor,
-        'sha': CharDescriptor,
         'binary': BinaryDescriptor,
         'selection': CharDescriptor,  # TODO implement its own descriptor
         'integer': IntegerDescriptor,
@@ -384,7 +389,7 @@ class MetaModelFactory(object):
         self.config = config or proteus.config.get_config()
 
     def __call__(self):
-        models_key = 'c%s' % id(self.config)
+        models_key = 'c%su%s' % (id(self.config), self.config.user)
         if not hasattr(_MODELS, models_key):
             setattr(_MODELS, models_key, {})
 
@@ -509,8 +514,8 @@ class ModelList(list):
     def pop(self, index=-1):
         self.record_removed.add(self[index])
         self[index]._parent = None
-        self[index]._parent_field_name = ''
-        self[index]._parent_name = ''
+        self[index]._parent_field_name = None
+        self[index]._parent_name = None
         res = super(ModelList, self).pop(index)
         self._changed()
         return res
@@ -519,8 +524,8 @@ class ModelList(list):
     def remove(self, record):
         self.record_deleted.add(record)
         record._parent = None
-        record._parent_field_name = ''
-        record._parent_name = ''
+        record._parent_field_name = None
+        record._parent_name = None
         res = super(ModelList, self).remove(record)
         self._changed()
         return res
@@ -534,11 +539,11 @@ class ModelList(list):
         raise NotImplementedError
     sort.__doc__ = list.sort.__doc__
 
-    def new(self):
+    def new(self, **kwargs):
         'Adds a new record to the ModelList and returns it'
         Relation = Model.get(self.model_name, self.parent._config)
         with Relation._config.set_context(self._get_context()):
-            new_record = Relation()
+            new_record = Relation(**kwargs)
         self.append(new_record)
         return new_record
 
@@ -612,7 +617,7 @@ class Model(object):
     def reset(cls, config=None, *names):
         'Reset class definition for Models named'
         config = config or proteus.config.get_config()
-        models_key = 'c%s' % id(config)
+        models_key = 'c%su%s' % (id(config), config.user)
         if not names:
             setattr(_MODELS, models_key, {})
         else:
@@ -683,6 +688,12 @@ class Model(object):
         self.reload()
         return True
 
+    def click(self, button):
+        'Click on button'
+        self.save()
+        self.reload()  # Force reload because save doesn't always
+        return getattr(self._proxy, button)([self.id], self._config.context)
+
     def _get_values(self, fields=None):
         'Return dictionary values'
         if fields is None:
@@ -760,11 +771,14 @@ class Model(object):
         return values
 
     def _get_on_change_value(self):
-        values = self._get_eval()
+        values = {'id': self.id}
         for field, definition in self._fields.iteritems():
-            if definition['type'] == 'one2many':
-                values[field] = [x._get_on_change_value()
-                    for x in getattr(self, field)]
+            if field in self._values and field != 'id':
+                if definition['type'] == 'one2many':
+                    values[field] = [x._get_on_change_value()
+                        for x in getattr(self, field)]
+                else:
+                    values[field] = getattr(self, '__%s_eval' % field)
         return values
 
     def _on_change_args(self, args):
@@ -777,10 +791,11 @@ class Model(object):
             scope = values
             for i in arg.split('.'):
                 if i not in scope:
-                    scope = False
                     break
                 scope = scope[i]
-            res[arg] = scope
+            else:
+                res[arg] = scope
+        res['id'] = self.id
         return res
 
     def _on_change_set(self, field, value):
@@ -795,8 +810,8 @@ class Model(object):
             for record in to_remove:
                 # remove without signal
                 list.remove(getattr(self, field), record)
-            if value and (value.get('add') or value.get('update')):
-                for vals in value.get('add', []):
+            if value and value.get('add') or value.get('update'):
+                for index, vals in value.get('add', []):
                     relation = Model.get(self._fields[field]['relation'],
                             self._config)
                     record = relation(_default=False)
@@ -806,7 +821,10 @@ class Model(object):
                         record._values[i] = j
                         record._changed.add(i)
                     # append without signal
-                    list.append(getattr(self, field), record)
+                    if index == -1:
+                        list.append(getattr(self, field), record)
+                    else:
+                        list.insert(getattr(self, field), index, record)
                 for vals in value.get('update', []):
                     if 'id' not in vals:
                         continue
@@ -878,7 +896,8 @@ class Model(object):
 class Wizard(object):
     'Wizard class for Tryton wizards'
 
-    def __init__(self, name, models=None, config=None, context=None):
+    def __init__(self, name, models=None, action=None, config=None,
+            context=None):
         if models:
             assert len(set(type(x) for x in models)) == 1
         super(Wizard, self).__init__()
@@ -892,6 +911,7 @@ class Wizard(object):
         self.session_id, self.start_state, self.end_state = result
         self.states = [self.start_state]
         self.models = models
+        self.action = action
         self.execute(self.start_state)
 
     def execute(self, state):
@@ -909,6 +929,10 @@ class Wizard(object):
                 ctx['active_id'] = None
                 ctx['active_ids'] = None
                 ctx['active_model'] = None
+            if self.action:
+                ctx['action_id'] = self.action.id
+            else:
+                ctx['action_id'] = None
 
             if self.form:
                 # Filter only modified values
